"use strict";
// tslint:disable-next-line:no-reference
/// <reference path="../../node_modules/reflect-metadata/index.d.ts" />
Object.defineProperty(exports, "__esModule", { value: true });
const graphql_1 = require("graphql");
const common_1 = require("./common");
const scalar_type_1 = require("./scalar-type");
const input_object_type_1 = require("./input-object-type");
const helpers_1 = require("../utils/helpers");
exports.GRAPHQL_OBJECT_TYPE_CONFIG = 'graphql:object-type-config';
exports.GRAPHQL_OBJECT_TYPE = 'graphql:object-type';
exports.GRAPHQL_OBJECT_TYPE_CONFIG_BUILD_QUEUE = 'graphql-object-type-build-queue';
exports.CONTEXT_INJECTOR_FACTORY = 'context-injector:factory';
exports.PROPERTY_KEYS = 'property-keys';
exports.CLASS_NAMES = 'class-names';
function Arg(argumentName, config) {
    return (target, propertyKey, parameterIndex) => {
        const existingGraphQLObjectTypeQueue = Reflect.getMetadata(exports.GRAPHQL_OBJECT_TYPE_CONFIG_BUILD_QUEUE, target.constructor, target.constructor.name) || [];
        existingGraphQLObjectTypeQueue.push(() => {
            const classNames = Reflect.getMetadata(exports.CLASS_NAMES, target.constructor);
            const classNameIndex = classNames.indexOf(target.constructor.name);
            let superClassConfig = {};
            if (classNameIndex > 0) {
                const superClassName = classNames[classNameIndex - 1];
                getObjectTypeFromClass(target.constructor, superClassName);
                superClassConfig = Reflect.getMetadata(exports.GRAPHQL_OBJECT_TYPE_CONFIG, target.constructor, superClassName);
            }
            const existingConfig = Reflect.getMetadata(exports.GRAPHQL_OBJECT_TYPE_CONFIG, target.constructor, target.constructor.name) || superClassConfig;
            const fieldName = propertyKey;
            existingConfig.fields = existingConfig.fields || {};
            existingConfig.fields[fieldName] = existingConfig.fields[fieldName] || {};
            existingConfig.fields[fieldName].args = existingConfig.fields[fieldName].args || {};
            existingConfig.fields[fieldName].args.__defineGetter__(argumentName, () => {
                const argumentType = (config && config.type) || Reflect.getMetadata(common_1.DESIGN_PARAMTYPES, target, propertyKey)[parameterIndex];
                const argumentGraphQLInputType = input_object_type_1.getInputTypeFromClass(argumentType) || scalar_type_1.getScalarTypeFromClass(argumentType) || argumentType;
                if (config && 'nullable' in config && !config.nullable) {
                    return {
                        type: new graphql_1.GraphQLNonNull(argumentGraphQLInputType)
                    };
                }
                else {
                    return {
                        type: argumentGraphQLInputType
                    };
                }
            });
            Reflect.defineMetadata(exports.GRAPHQL_OBJECT_TYPE_CONFIG, existingConfig, target.constructor, target.constructor.name);
        });
        Reflect.defineMetadata(exports.GRAPHQL_OBJECT_TYPE_CONFIG_BUILD_QUEUE, existingGraphQLObjectTypeQueue, target.constructor, target.constructor.name);
    };
}
exports.Arg = Arg;
function Field(typeFactory, config) {
    return (target, propertyKey) => {
        const existingGraphQLObjectTypeQueue = Reflect.getMetadata(exports.GRAPHQL_OBJECT_TYPE_CONFIG_BUILD_QUEUE, target.constructor, target.constructor.name) || [];
        existingGraphQLObjectTypeQueue.push(() => {
            const classNames = Reflect.getMetadata(exports.CLASS_NAMES, target.constructor);
            const classNameIndex = classNames.indexOf(target.constructor.name);
            let superClassConfig = {};
            if (classNameIndex > 0) {
                const superClassName = classNames[classNameIndex - 1];
                getObjectTypeFromClass(target.constructor, superClassName);
                superClassConfig = Reflect.getMetadata(exports.GRAPHQL_OBJECT_TYPE_CONFIG, target.constructor, superClassName);
            }
            const existingConfig = Reflect.getMetadata(exports.GRAPHQL_OBJECT_TYPE_CONFIG, target.constructor, target.constructor.name) || superClassConfig;
            const fieldName = (config && config.name) || propertyKey;
            existingConfig.fields = existingConfig.fields || {};
            existingConfig.fields[fieldName] = existingConfig.fields[fieldName] || {};
            existingConfig.fields[fieldName].__defineGetter__('type', () => {
                const fieldType = typeFactory ? typeFactory() : (typeof target[propertyKey] === 'function' ? Reflect.getMetadata(common_1.DESIGN_RETURNTYPE, target, propertyKey) : Reflect.getMetadata(common_1.DESIGN_TYPE, target, propertyKey));
                const graphQLType = getObjectTypeFromClass(fieldType) || scalar_type_1.getScalarTypeFromClass(fieldType) || fieldType;
                if (config && 'nullable' in config && !config.nullable) {
                    return new graphql_1.GraphQLNonNull(graphQLType);
                }
                else {
                    return graphQLType;
                }
            });
            if (typeof target[propertyKey] === 'function') {
                existingConfig.fields[fieldName].resolve = ((root, args, context) => {
                    // If 3rd party DI container is defined
                    if (Reflect.getMetadata(exports.CONTEXT_INJECTOR_FACTORY, target.constructor, target.constructor.name)) {
                        const contextInjectorFactoryDefinition = Reflect.getMetadata(exports.CONTEXT_INJECTOR_FACTORY, target.constructor, target.constructor.name);
                        let injector;
                        if (typeof contextInjectorFactoryDefinition === 'function') {
                            injector = contextInjectorFactoryDefinition(context);
                        }
                        else if (typeof contextInjectorFactoryDefinition === 'object') {
                            injector = contextInjectorFactoryDefinition;
                        }
                        const keys = Reflect.getMetadata(exports.PROPERTY_KEYS, target.constructor, target.constructor.name) || [];
                        for (const key of keys) {
                            if (Reflect.getMetadata(common_1.DESIGN_TYPE, target, key)) {
                                root = root || {};
                                Object.defineProperty(root, key, {
                                    get() {
                                        const serviceIdentifier = Reflect.getMetadata(common_1.DESIGN_TYPE, target, key);
                                        return injector.get(serviceIdentifier);
                                    }
                                });
                            }
                        }
                    }
                    return target[propertyKey].call(root, ...Object['values'](args));
                }); // TODO: NOT SAFE
            }
            Reflect.defineMetadata(exports.GRAPHQL_OBJECT_TYPE_CONFIG, existingConfig, target.constructor, target.constructor.name);
        });
        Reflect.defineMetadata(exports.GRAPHQL_OBJECT_TYPE_CONFIG_BUILD_QUEUE, existingGraphQLObjectTypeQueue, target.constructor, target.constructor.name);
    };
}
exports.Field = Field;
function Inject(serviceIdentifier) {
    return (target, propertyKey, index) => {
        const allDependencies = Reflect.getMetadata(common_1.DESIGN_PARAMTYPES, target) || [];
        const propertyKeys = Reflect.getMetadata(exports.PROPERTY_KEYS, target.constructor || target, (target.constructor || target).name) || [];
        if (typeof propertyKey === 'undefined') {
            if (typeof index !== 'undefined') {
                allDependencies[index] = serviceIdentifier;
            }
        }
        else {
            const designType = serviceIdentifier || Reflect.getMetadata(common_1.DESIGN_TYPE, target, propertyKey);
            Reflect.defineMetadata(common_1.DESIGN_TYPE, designType, target, propertyKey);
            propertyKeys.push(propertyKey);
        }
        Reflect.defineMetadata(common_1.DESIGN_PARAMTYPES, allDependencies, target);
        Reflect.defineMetadata(exports.PROPERTY_KEYS, propertyKeys, target.constructor || target, (target.constructor || target).name);
        return target;
    };
}
exports.Inject = Inject;
function ObjectType(config = {}) {
    return target => {
        const classNames = Reflect.getMetadata(exports.CLASS_NAMES, target) || [];
        classNames.push(target.name);
        Reflect.defineMetadata(exports.CLASS_NAMES, classNames, target);
        // Delete the existing metadata on redeclaration, because it should override the existing generated ObjectType of the inherited super class
        Reflect.defineMetadata(exports.GRAPHQL_OBJECT_TYPE, undefined, target, target.name);
        if (config.injector) {
            Reflect.defineMetadata(exports.CONTEXT_INJECTOR_FACTORY, config.injector, target, target.name);
        }
        const existingGraphQLObjectTypeQueue = Reflect.getMetadata(exports.GRAPHQL_OBJECT_TYPE_CONFIG_BUILD_QUEUE, target, target.name) || [];
        existingGraphQLObjectTypeQueue.push(() => {
            const classNames = Reflect.getMetadata(exports.CLASS_NAMES, target);
            const classNameIndex = classNames.indexOf(target.name);
            let superClassConfig = {};
            if (classNameIndex > 0) {
                const superClassName = classNames[classNameIndex - 1];
                getObjectTypeFromClass(target, superClassName);
                superClassConfig = Reflect.getMetadata(exports.GRAPHQL_OBJECT_TYPE_CONFIG, target, superClassName);
            }
            const existingConfig = Reflect.getMetadata(exports.GRAPHQL_OBJECT_TYPE_CONFIG, target, target.name) || superClassConfig;
            Reflect.defineMetadata(exports.GRAPHQL_OBJECT_TYPE, new graphql_1.GraphQLObjectType(Object.assign({ name: config.name || target.name, interfaces: helpers_1.asArray(config.implements).map(interfaceType => getObjectTypeFromClass(interfaceType) || interfaceType) }, existingConfig)), target, target.name);
        });
        Reflect.defineMetadata(exports.GRAPHQL_OBJECT_TYPE_CONFIG_BUILD_QUEUE, existingGraphQLObjectTypeQueue, target, target.name);
        return target;
    };
}
exports.ObjectType = ObjectType;
function getObjectTypeFromClass(target, name = target.name) {
    if (target instanceof Array) {
        const elementType = getObjectTypeFromClass(target[0]);
        return elementType && new graphql_1.GraphQLList(elementType);
    }
    if (!Reflect.getMetadata(exports.GRAPHQL_OBJECT_TYPE, target, name) &&
        Reflect.getMetadata(exports.GRAPHQL_OBJECT_TYPE_CONFIG_BUILD_QUEUE, target, name)) {
        const existingGraphQLObjectTypeQueue = Reflect.getMetadata(exports.GRAPHQL_OBJECT_TYPE_CONFIG_BUILD_QUEUE, target, name);
        existingGraphQLObjectTypeQueue.forEach(fn => fn());
    }
    return Reflect.getMetadata(exports.GRAPHQL_OBJECT_TYPE, target, name);
}
exports.getObjectTypeFromClass = getObjectTypeFromClass;
//# sourceMappingURL=object-type.js.map